from qtasync.qasyncio._env import QtEventLoop

try:
    from qtasync.qasyncio._env import QtProactorEventLoop
except ImportError:
    QtProactorEventLoop = None

try:
    from qtasync.qasyncio._env import QtSelectorEventLoop
except ImportError:
    QtSelectorEventLoop = None

__all__ = ["QtEventLoop", "QtProactorEventLoop", "QtSelectorEventLoop"]



import threading
from typing import TYPE_CHECKING, Optional

from qtasync._env import QThread
from qtasync._util import qt_timeout
from qtasync.types.bound import PYTHON_TIME

if TYPE_CHECKING:
    from qtasync._env import QObject


class QtThread(QThread):
    def __init__(
        self,
        target=None,
        name=None,
        args: list = None,
        kwargs: dict = None,
        *,
        daemon=None,
        parent: "QObject" = None
    ):
        super().__init__(parent=parent)
        self._fn = target
        self._name = name
        self._args = args or []
        self._kwargs = kwargs or {}
        self._daemon = daemon
        self._py_thread: Optional["threading.Thread"] = None

    def run(self):
        self._py_thread = threading.current_thread()
        self._py_thread.setName(self._name)
        self._fn(*self._args, **self._kwargs)
        self._py_thread._is_stopped = True

    def join(self, timeout: PYTHON_TIME = None):
        self.wait(qt_timeout(timeout))

    @property
    def ident(self) -> Optional[int]:
        return None if not self._py_thread else self._py_thread.ident

    def is_alive(self) -> bool:
        return self.isRunning()



from typing import TYPE_CHECKING, Callable, Any

from qtasync._env import QTimer
from qtasync._util import qt_timeout
from qtasync.types.unbound import SIGNAL_TYPE
from qtasync.types.bound import PYTHON_TIME

if TYPE_CHECKING:
    from qtasync._env import QObject


class QtTimer(QTimer):
    timeout: SIGNAL_TYPE

    def cancel(self):
        self.stop()

    @classmethod
    def singleShot(
        cls, duration: PYTHON_TIME, func: Callable[[], Any], parent: "QObject" = None
    ) -> "QtTimer":
        t = cls(parent=parent)
        t.setSingleShot(True)
        t.setInterval(qt_timeout(duration))
        t.timeout.connect(func)
        t.start()
        return t



import logging
import os
import threading
import sys
from enum import Enum
from typing import Callable, Any, Optional
from concurrent.futures import Executor, Future, CancelledError, InvalidStateError
from concurrent.futures import TimeoutError as FutureTimeoutError
from concurrent.futures._base import (
    PENDING,
    RUNNING,
    CANCELLED,
    CANCELLED_AND_NOTIFIED,
    FINISHED,
)

from qtasync._env import (
    QObject,
    Signal,
    QThreadPool,
    QRunnable,
)

from qtasync.qthreading import QtLock, QtCondition
from qtasync.types.bound import PYTHON_TIME
from qtasync.types.unbound import SIGNAL_TYPE

log = logging.getLogger(__name__)


class FutureStatus(Enum):
    PENDING = PENDING
    RUNNING = RUNNING
    CANCELLED = CANCELLED
    CANCELLED_AND_NOTIFIED = CANCELLED_AND_NOTIFIED
    FINISHED = FINISHED


class _QRunnable(QRunnable):
    def __init__(self, future: "QtFuture", fn: Callable, args, kwargs):
        super().__init__()
        self.future = future
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def run(self):
        # Get coverage on the runnable for tests, courtesy of @earonesty
        # https://github.com/nedbat/coveragepy/issues/686#issuecomment-631760644
        if threading._trace_hook:
            sys.settrace(threading._trace_hook)
        if not self.future.set_running_or_notify_cancel():
            return

        try:
            log.debug("Running %s", self)
            result = self.fn(*self.args, **self.kwargs)
            log.debug("Result was %s", result)
        except BaseException as ex:
            self.future.set_exception(ex)
            # Copied from concurrent.futures.thread._WorkItem.run()
            # self = None
        else:
            self.future.set_result(result)


class QtFuture(QObject, Future):
    """
    This class is unique in that it not only implements the interface of
    concurrent.futures.Future, but it also subclasses it. This is due to
    the type requirement of asyncio.BaseEventLoop.run_in_executor() that the
    return value of executor.submit() be an instance of concurrent.futures.Future.
    """

    _finished: SIGNAL_TYPE = Signal()

    def __init__(self, parent: "QObject" = None):
        super().__init__(parent=parent)
        self._id = os.urandom(8).hex()

        self._result: Optional[Any] = None
        self._exception: Optional[BaseException] = None

        self._cond = QtCondition()
        self._state: "FutureStatus" = FutureStatus.PENDING

    def __repr__(self):
        with self._cond:
            if self._state == FINISHED:
                if self._exception:
                    return "<%s at %#x state=%s raised %s>" % (
                        self.__class__.__name__,
                        id(self),
                        self._state.value,
                        self._exception.__class__.__name__,
                    )
                else:
                    return "<%s at %#x state=%s returned %s>" % (
                        self.__class__.__name__,
                        id(self),
                        self._state.value,
                        self._result.__class__.__name__,
                    )
            return "<%s at %#x state=%s>" % (
                self.__class__.__name__,
                id(self),
                self._state.value,
            )

    @property
    def future_id(self) -> str:
        return self._id

    def cancel(self) -> bool:
        with self._cond:
            if self._state in [FutureStatus.FINISHED, FutureStatus.RUNNING]:
                return False
            elif self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]:
                return True

            self._state = FutureStatus.CANCELLED
            self._cond.notify_all()

    def cancelled(self) -> bool:
        with self._cond:
            return self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]

    def running(self) -> bool:
        with self._cond:
            return self._state == FutureStatus.RUNNING

    def done(self) -> bool:
        with self._cond:
            return self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
                FutureStatus.FINISHED,
            ]

    def __get_result(self):
        if self._exception:
            raise self._exception
        else:
            return self._result

    def result(self, timeout: Optional[PYTHON_TIME] = None):
        with self._cond:
            if self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]:
                raise CancelledError()
            elif self._state == FutureStatus.FINISHED:
                return self.__get_result()

            if not self._cond.wait(timeout=timeout):
                raise FutureTimeoutError

            if self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]:
                raise CancelledError()
            elif self._state == FutureStatus.FINISHED:
                return self.__get_result()
            else:
                raise FutureTimeoutError()

    def exception(
        self, timeout: Optional[PYTHON_TIME] = None
    ) -> Optional[BaseException]:
        with self._cond:
            if self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]:
                raise CancelledError()
            elif self._state == FutureStatus.FINISHED:
                return self._exception

            self._cond.wait(timeout=timeout)

            if self._state in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
            ]:
                raise CancelledError()
            elif self._state == FutureStatus.FINISHED:
                return self._exception
            else:
                raise TimeoutError()

    def add_done_callback(self, fn: Callable[["QtFuture"], Any]) -> None:
        with self._cond:
            if self._state not in [
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
                FutureStatus.FINISHED,
            ]:
                self._finished.connect(lambda: fn(self))
                return

        try:
            fn(self)
        except:  # noqa: E722
            log.exception("Error when calling PythonicQFuture done callback")

    # Testing and Executor usage
    def set_running_or_notify_cancel(self) -> bool:
        with self._cond:
            if self._state == FutureStatus.CANCELLED:
                self._state = FutureStatus.CANCELLED_AND_NOTIFIED
                return False
            elif self._state == FutureStatus.PENDING:
                self._state = FutureStatus.RUNNING
                return True
            else:
                log.critical("Future %s in unexpected state: %s", id(self), self._state)
                raise RuntimeError("Future in unexpected state")

    def set_result(self, result) -> None:
        with self._cond:
            if self._state in (
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
                FutureStatus.FINISHED,
            ):
                raise InvalidStateError("{}: {!r}".format(self._state, self))
            self._result = result
            self._state = FutureStatus.FINISHED
            self._cond.notify_all()
            self._finished.emit()

    def set_exception(self, exception: Optional[BaseException]) -> None:
        with self._cond:
            if self._state in (
                FutureStatus.CANCELLED,
                FutureStatus.CANCELLED_AND_NOTIFIED,
                FutureStatus.FINISHED,
            ):
                raise InvalidStateError("{}: {!r}".format(self._state, self))
            self._exception = exception
            self._state = FutureStatus.FINISHED
            self._cond.notify_all()
            self._finished.emit()


class QtThreadPoolExecutor(Executor):
    def __init__(self, qthread_pool: "QThreadPool" = None):
        self._pool = qthread_pool or QThreadPool.globalInstance()
        self._shutdown_mutex = QtLock()
        self._is_shutdown = False

    def submit(self, fn, *args, **kwargs) -> QtFuture:
        with self._shutdown_mutex:
            if self._is_shutdown:
                raise RuntimeError
            future = QtFuture(parent=self._pool)
            runnable = _QRunnable(future, fn, args, kwargs)
            log.debug("Submitting to QThreadPoolExecutor: %s(%s, %s)", fn, args, kwargs)
            self._pool.start(runnable)
            return future

    def shutdown(self, wait=True, *, cancel_futures=False):
        with self._shutdown_mutex:
            self._is_shutdown = True

            if cancel_futures:
                self._pool.clear()

        if wait:
            self._pool.waitForDone()

    def __enter__(self):
        with self._shutdown_mutex:
            if self._is_shutdown:
                raise RuntimeError("QThreadPoolExecutor has been shut down already")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()



import logging
import importlib
import sys
import os
from typing import TYPE_CHECKING, Type, Optional

from qtasync.types.unbound import SIGNAL_TYPE

if TYPE_CHECKING:
    # For static type checking, feel free to set the Qt library of your choice here
    from PyQt5 import QtCore as _TypedQtCore
    from PyQt5 import QtWidgets as _TypedQtWidgets


_log = logging.getLogger(__name__)

_QtModule = None
QtModuleName = None
PYSIDE2_MODULE_NAME = "PySide2"
PYSIDE6_MODULE_NAME = "PySide6"
PYQT5_MODULE_NAME = "PyQt5"
PYQT6_MODULE_NAME = "PyQt6"

# If QT_API env variable is given, use that or fail trying
_qtapi_env = os.getenv("QT_API", "").strip().lower()
if _qtapi_env:
    env_to_mod_map = {
        "pyqt5": PYQT5_MODULE_NAME,
        "pyqt6": PYQT6_MODULE_NAME,
        "pyside6": PYSIDE6_MODULE_NAME,
        "pyside2": PYSIDE2_MODULE_NAME,
    }
    if _qtapi_env in env_to_mod_map:
        QtModuleName = env_to_mod_map[_qtapi_env]
    else:
        raise ImportError(
            "QT_API environment variable set ({}) but not one of [{}].".format(
                _qtapi_env, ", ".join(env_to_mod_map.keys())
            )
        )

    _log.info("Forcing use of {} as Qt Implementation".format(QtModuleName))
    _QtModule = importlib.import_module(QtModuleName)

# If a Qt lib is already imported, use that
if not _QtModule:
    for QtModuleName in ("PySide2", "PySide6", "PyQt5", "PyQt6"):
        if QtModuleName in sys.modules:
            _QtModule = sys.modules[QtModuleName]
            break

# Try importing qt libs
if not _QtModule:
    for QtModuleName in ("PySide2", "PySide6", "PyQt5", "PyQt6"):
        try:
            _QtModule = importlib.import_module(QtModuleName)
        except ImportError:
            continue
        else:
            break


if not _QtModule:
    raise ImportError("No Qt implementations found")

_log.info("Using Qt Implementation: {}".format(QtModuleName))

if QtModuleName == "PyQt5":
    from PyQt5 import QtCore as _QtCore
    from PyQt5 import QtWidgets as _QtWidgets
elif QtModuleName == "PyQt6":
    from PyQt6 import QtCore as _QtCore
    from PyQt6 import QtWidgets as _QtWidgets
elif QtModuleName == "PySide2":
    from PySide2 import QtCore as _QtCore
    from PySide2 import QtWidgets as _QtWidgets
elif QtModuleName == "PySide6":
    from PySide6 import QtCore as _QtCore
    from PySide6 import QtWidgets as _QtWidgets
else:
    raise ImportError("Failed to import QCoreApplication")

# Expose Qt components and modules for importation
# QtCore
QtCore: "_TypedQtCore" = _QtCore
QCoreApplication: Type["_TypedQtCore.QCoreApplication"] = _QtCore.QCoreApplication
try:
    Slot: "_TypedQtCore.Slot" = _QtCore.Slot
except AttributeError:
    Slot: "_TypedQtCore.Slot" = _QtCore.pyqtSlot
try:
    Signal: "_TypedQtCore.Signal" = _QtCore.Signal
except AttributeError:
    Signal: "_TypedQtCore.pyqtSignal" = _QtCore.pyqtSignal
try:
    SignalInstance: Type["_TypedQtCore.SignalInstance"] = _QtCore.SignalInstance
except AttributeError:
    SignalInstance: Type["_TypedQtCore.pyqtBoundSignal"] = _QtCore.pyqtBoundSignal
QObject: Type["_TypedQtCore.QObject"] = _QtCore.QObject
QSocketNotifier: Type["_TypedQtCore.QSocketNotifier"] = _QtCore.QSocketNotifier
QMutexLocker: Type["_TypedQtCore.QMutexLocker"] = _QtCore.QMutexLocker
QThread: Type["_TypedQtCore.QThread"] = _QtCore.QThread
QSemaphore: Type["_TypedQtCore.QSemaphore"] = _QtCore.QSemaphore
QMutex: Type["_TypedQtCore.QMutex"] = _QtCore.QMutex
if QtModuleName in (PYQT6_MODULE_NAME, PYSIDE6_MODULE_NAME):
    QRecursiveMutex: Optional[
        Type["_TypedQtCore.QRecursiveMutex"]
    ] = _QtCore.QRecursiveMutex
else:
    QRecursiveMutex = None
QWaitCondition: Type["_TypedQtCore.QWaitCondition"] = _QtCore.QWaitCondition
QTimer: Type["_TypedQtCore.QTimer"] = _QtCore.QTimer
QThreadPool: Type["_TypedQtCore.QThreadPool"] = _QtCore.QThreadPool
QRunnable: Type["_TypedQtCore.QRunnable"] = _QtCore.QRunnable
QEvent: Type["_TypedQtCore.QEvent"] = _QtCore.QEvent
QDeadlineTimer: Type["_TypedQtCore.QDeadlineTimer"] = _QtCore.QDeadlineTimer
QTimerEvent: Type["_TypedQtCore.QTimerEvent"] = _QtCore.QTimerEvent
QTime: Type["_TypedQtCore.QTime"] = _QtCore.QTime
QElapsedTimer: Type["_TypedQtCore.QElapsedTimer"] = _QtCore.QElapsedTimer
QtDebugMsg: Type["_TypedQtCore.QtMsgType.QtDebugMsg"] = _QtCore.QtMsgType.QtDebugMsg
QtInfoMsg: Type["_TypedQtCore.QtMsgType.QtInfoMsg"] = _QtCore.QtMsgType.QtInfoMsg
QtWarningMsg: Type[
    "_TypedQtCore.QtMsgType.QtWarningMsg"
] = _QtCore.QtMsgType.QtWarningMsg
QtCriticalMsg: Type[
    "_TypedQtCore.QtMsgType.QtCriticalMsg"
] = _QtCore.QtMsgType.QtCriticalMsg
QtFatalMsg: Type["_TypedQtCore.QtMsgType.QtFatalMsg"] = _QtCore.QtMsgType.QtFatalMsg


# QtWidgets
QApplication: Type["_TypedQtWidgets.QApplication"] = _QtWidgets.QApplication

if TYPE_CHECKING:
    # Subclass of whatever QApplication is with some type hints and implementation-independent shims
    # But this is never actually created, as there is a bug where subclasses of a QCoreApp will cause segfaults
    # on exit: https://bugreports.qt.io/browse/PYSIDE-1447
    class QCoreApplication(QCoreApplication):
        aboutToQuit: SIGNAL_TYPE
        applicationNameChanged: SIGNAL_TYPE
        applicationVersionChanged: SIGNAL_TYPE
        organizationDomainChanged: SIGNAL_TYPE
        organizationNameChanged: SIGNAL_TYPE

        def send_posted_events(
            self: "_TypedQtCore.QCoreApplication",
            receiver: "QObject" = None,
            event_type=0,
        ):
            ...


def send_posted_events(
    self: "_TypedQtCore.QCoreApplication", receiver: "QObject" = None, event_type=0
):
    if QtModuleName in (PYQT5_MODULE_NAME, PYQT6_MODULE_NAME):
        return self.sendPostedEvents(receiver=receiver, eventType=event_type)
    else:
        return self.sendPostedEvents(receiver=receiver, event_type=event_type)


# Enable this function to view all events processed in the qapp
if os.environ.get("LOG_QT_EVENTS", False):
    import traceback

    orig_notify = QCoreApplication.notify

    def notify(self: "_TypedQtCore.QCoreApplication", obj: "QObject", event: "QEvent"):
        ret = orig_notify(self, obj, event)
        try:
            _log.debug("Notify %s %s (%s) returned %s", obj, event, event.type(), ret)
            tb = traceback.extract_stack()
            _log.debug("".join(tb.format()))
        except:  # noqa: E722
            _log.warning(
                "Failed to analyze QCoreApp event (%s, %s, %s)", self, obj, event
            )
        finally:
            return ret

    QCoreApplication.notify = notify
QCoreApplication.send_posted_events = send_posted_events



import logging
from typing import Optional, Union, Callable, Any

from qtasync._env import (
    QSemaphore,
    QMutex,
    QRecursiveMutex,
    QWaitCondition,
    QtModuleName,
    PYQT5_MODULE_NAME,
    PYQT6_MODULE_NAME,
    PYSIDE6_MODULE_NAME,
    QElapsedTimer,
)

from qtasync.types.bound import QT_TIME, PYTHON_TIME
from qtasync._util import qt_timeout, mk_q_deadline_timer, py_timeout

log = logging.getLogger(__name__)


def _get_ident() -> int:
    import threading

    try:
        return threading.get_native_id()
    except:  # noqa
        return threading.get_ident()


class _QtLock:
    def __init__(self, default_timeout: PYTHON_TIME = -1, recursive: bool = False):
        """
        :param default_timeout: A divergence from Python's Lock/RLock, this parameter can
            be used to set a timeout used by the mutex when acquired in a `with` block (or
            if you explicitly call __enter__())
        :param recursive: Whether or not the mutex can be re-acquired by the same thread
        """
        if QtModuleName in (PYQT6_MODULE_NAME, PYSIDE6_MODULE_NAME):
            self._mutex = QRecursiveMutex() if recursive else QMutex()
        else:
            self._mutex = QMutex(QMutex.Recursive if recursive else QMutex.NonRecursive)
        self._default_timeout: QT_TIME = qt_timeout(default_timeout)

    # Python methods to match threading.Lock/RLock's interface
    def __enter__(self):
        if not self._try_lock(timeout=self._default_timeout):
            raise TimeoutError("QMutex timed out")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def acquire(self, blocking=True, timeout: PYTHON_TIME = -1.0):
        if blocking:
            # Negative timeout in a QMutex behaves the same as a Python lock
            return self._try_lock(timeout=qt_timeout(timeout))
        else:
            if timeout != -1.0:
                raise ValueError("Cannot specify a timeout for a non-blocking call")
            return self._try_lock()

    def release(self):
        self._mutex.unlock()

    @property
    def default_timeout(self) -> float:
        return self._default_timeout

    @default_timeout.setter
    def default_timeout(self, new_timeout: float):
        self._default_timeout = qt_timeout(new_timeout)

    def is_recursive(self) -> bool:
        return self._mutex.isRecursive()

    def _try_lock(self, timeout: QT_TIME = None) -> bool:
        if timeout is None:
            if QtModuleName in (PYQT5_MODULE_NAME, PYQT6_MODULE_NAME):
                return self._mutex.tryLock()
            else:
                return self._mutex.try_lock()
        else:
            return self._mutex.tryLock(timeout)

    def _is_owned(self) -> bool:
        raise NotImplementedError


class QtRLock(_QtLock):
    def __init__(self, default_timeout: PYTHON_TIME = -1):
        super().__init__(default_timeout=default_timeout, recursive=True)
        self._owner = None
        self._count = 0

    def _try_lock(self, timeout: QT_TIME = None) -> bool:
        me = _get_ident()
        if self._owner == me:
            self._count += 1
            return True
        ret = super()._try_lock(timeout=timeout)
        if ret:
            self._owner = me
            self._count = 1
        return ret

    def release(self):
        if self._owner != _get_ident():
            raise RuntimeError("Cannot release un-acquired QtRLock")
        self._count -= 1
        if self._count == 0:
            self._owner = None
            super().release()

    def _recursion_count(self) -> int:
        return self._count

    def _is_owned(self) -> bool:
        return _get_ident() == self._owner


class QtLock(_QtLock):
    def __init__(self, default_timeout: PYTHON_TIME = -1):
        super().__init__(default_timeout=default_timeout, recursive=False)

    def _is_owned(self):
        if self.acquire(blocking=False):
            self.release()
            return False
        else:
            return True


class QtCondition:
    def __init__(self, lock: Union["QtLock", "QtRLock"] = None):
        self._mutex = lock or QtLock()
        # Cannot use a recursive mutex in Qt 5
        # See: https://github.com/qt/qtbase/blob/5.15.2/src/corelib/thread/qwaitcondition_unix.cpp#L217
        #  and https://github.com/qt/qtbase/blob/5.15.2/src/corelib/thread/qwaitcondition_win.cpp#L171
        # TODO: If using Qt 6, allow recursive mutexes as long as the mutex is not recursively locked
        #   see: https://github.com/qt/qtbase/blob/6.3/src/corelib/thread/qwaitcondition_win.cpp#L189
        assert not isinstance(self._mutex, QtRLock)
        self._cond = QWaitCondition()

    # Python methods to match threading.Condition
    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def acquire(self, blocking=True, timeout: PYTHON_TIME = -1.0):
        return self._mutex.acquire(blocking=blocking, timeout=timeout)

    def release(self):
        if not self._mutex._is_owned():
            raise RuntimeError("Cannot release un-acquired lock")
        self._mutex.release()

    def wait(self, timeout: PYTHON_TIME = None) -> bool:
        # Since the underlying mutex is not recursive, we can ensure that the mutex is locked by simply attempting to
        # lock it without blocking
        if not self._mutex._is_owned():
            raise RuntimeError("Cannot wait on un-acquired lock")

        if timeout is None:
            return self._cond.wait(self._mutex._mutex)
        else:
            if QtModuleName == PYQT5_MODULE_NAME:
                # For some reason, the QDeadlineTimer does not work with PyQt5 and wait() instantly returns
                return self._cond.wait(self._mutex._mutex, msecs=qt_timeout(timeout))
            else:
                return self._cond.wait(self._mutex._mutex, mk_q_deadline_timer(timeout))

    def wait_for(self, predicate: Callable[[], Any], timeout: PYTHON_TIME = None):
        """
        Largely a copy of threading.Condition.wait_for()
        """
        timer = None
        waittime = qt_timeout(timeout)
        result = predicate()
        while not result:
            if waittime is not None:
                if timer is None:
                    timer = QElapsedTimer()
                    timer.start()
                else:
                    # New waittime is remaining wait time minus the time elapsed since the timer was last (re)started
                    waittime -= timer.restart()
                    if waittime <= 0:
                        break
            self.wait(py_timeout(waittime))
            result = predicate()
        return result

    def notify_all(self):
        if not self._mutex._is_owned():
            raise RuntimeError("Cannot notify all on un-acquired lock")
        self._cond.wakeAll()

    def notify(self):
        if not self._mutex._is_owned():
            raise RuntimeError("Cannot notify on un-acquired lock")
        self._cond.wakeOne()


class QtEvent:
    def __init__(self):
        self._is_set = False
        self._flag_mutex = QtLock()
        self._cond = QtCondition()

    def is_set(self) -> bool:
        with self._flag_mutex:
            return self._is_set

    def set(self):
        with self._flag_mutex:
            if not self._is_set:
                self._is_set = True
                with self._cond:
                    self._cond.notify_all()

    def clear(self):
        with self._flag_mutex:
            self._is_set = False

    def wait(self, timeout: Optional[PYTHON_TIME] = None) -> bool:
        """
        This wait function is designed to emulate threading.Event.wait()

        :param timeout: The time to wait before timing out in seconds
        """
        # Lock the event flag first
        with self._flag_mutex:
            if self._is_set:
                return True

        with self._cond:
            return self._cond.wait(timeout)


class QtSemaphore(QSemaphore):
    def __init__(self, value=1):
        super().__init__(n=value)

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def release(self, n=1):
        return super().release(n=n)

    def acquire(self, blocking=True, timeout: PYTHON_TIME = None) -> bool:
        if blocking:
            return self.tryAcquire(1, qt_timeout(timeout))
        else:
            return self.tryAcquire(1)



import os
import asyncio
import contextlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qtasync.qasyncio._loop import _QEventLoop


if os.name == "nt":
    from qtasync.qasyncio._windows import QtProactorEventLoop

    QtEventLoop = QtProactorEventLoop
else:
    from qtasync.qasyncio._unix import QtSelectorEventLoop  # noqa

    QtEventLoop = QtSelectorEventLoop


class QtEventLoopPolicyMixin(asyncio.DefaultEventLoopPolicy):
    def new_event_loop(self) -> "_QEventLoop":
        return QtEventLoop()


@contextlib.contextmanager
def _set_event_loop_policy(policy):
    old_policy = asyncio.get_event_loop_policy()
    asyncio.set_event_loop_policy(policy)
    try:
        yield
    finally:
        asyncio.set_event_loop_policy(old_policy)


def run(*args, **kwargs):
    with _set_event_loop_policy(QtEventLoopPolicyMixin()):
        return asyncio.run(*args, **kwargs)



import logging
import asyncio
import sys
from typing import TYPE_CHECKING

try:
    import _winapi
    from asyncio import windows_events
    import _overlapped
except ImportError:  # noqa
    pass  # w/o guarding this import py.test can't gather doctests on platforms w/o _winapi

import math

from qtasync.qasyncio._util import _make_signaller
from qtasync.qasyncio._loop import _QEventLoop
from qtasync._env import QMutex, QMutexLocker, QThread, QSemaphore

if TYPE_CHECKING:
    from qtasync._env import QObject


log = logging.getLogger(__name__)


UINT32_MAX = 0xFFFFFFFF


class QtProactorEventLoop(_QEventLoop, asyncio.ProactorEventLoop):

    """Proactor based event loop."""

    def __init__(self):
        self._proactor = _IocpProactor()
        super().__init__(self._proactor)

        self.__event_signaller = _make_signaller("QVariantList")
        self.__event_signal = self.__event_signaller.signal
        self.__event_signal.connect(self._process_events)
        self.__event_poller = _EventPoller(
            self.__event_signal, rsc_parent=self._rsc_parent
        )

    def _process_events(self, events):
        """Process events from proactor."""
        for f, callback, transferred, key, ov in events:
            try:
                log.debug("Invoking event callback %s", callback)
                value = callback(transferred, key, ov)
            except OSError as e:
                log.debug("Event callback failed", exc_info=sys.exc_info())
                if not f.done():
                    f.set_exception(e)
            else:
                if not f.cancelled():
                    f.set_result(value)

    def _check_closed(self):
        # TODO: Needed? Should be self._proactor._check_closed()?
        if not self._proactor:
            raise RuntimeError
        return self._proactor._check_closed()

    def _before_run_forever(self):
        self.__event_poller.start(self._proactor)

    def _after_run_forever(self):
        self.__event_poller.stop()


class _IocpProactor(windows_events.IocpProactor):
    def __init__(self):
        self.__events = []
        super().__init__()
        self._lock = QMutex()

    def select(self, timeout=None):
        """Override in order to handle events in a threadsafe manner."""
        if not self.__events:
            self._poll(timeout)
        tmp = self.__events
        self.__events = []
        return tmp

    def recv(self, conn, nbytes, flags=0):
        with QMutexLocker(self._lock):
            return super().recv(conn, nbytes, flags)

    def send(self, conn, buf, flags=0):
        with QMutexLocker(self._lock):
            return super().send(conn, buf, flags)

    def _poll(self, timeout=None):  # noqa: C901
        """Override in order to handle events in a threadsafe manner."""
        if timeout is None:
            ms = UINT32_MAX  # wait for eternity
        elif timeout < 0:
            raise ValueError("negative timeout")
        else:
            # GetQueuedCompletionStatus() has a resolution of 1 millisecond,
            # round away from zero to wait *at least* timeout seconds.
            ms = math.ceil(timeout * 1e3)
            if ms >= UINT32_MAX:
                raise ValueError("timeout too big")

        with QMutexLocker(self._lock):
            while True:
                # log.debug('Polling IOCP with timeout {} ms in thread {}...'.format(
                #     ms, threading.get_ident()))
                status = _overlapped.GetQueuedCompletionStatus(self._iocp, ms)
                if status is None:
                    break

                err, transferred, key, address = status
                try:
                    f, ov, obj, callback = self._cache.pop(address)
                except KeyError:
                    # key is either zero, or it is used to return a pipe
                    # handle which should be closed to avoid a leak.
                    if key not in (0, _overlapped.INVALID_HANDLE_VALUE):
                        _winapi.CloseHandle(key)
                    ms = 0
                    continue

                if obj in self._stopped_serving:
                    f.cancel()
                # Futures might already be resolved or cancelled
                elif not f.done():
                    self.__events.append((f, callback, transferred, key, ov))

                ms = 0

    def _wait_for_handle(self, handle, timeout, _is_cancel):
        with QMutexLocker(self._lock):
            return super()._wait_for_handle(handle, timeout, _is_cancel)

    def accept(self, listener):
        with QMutexLocker(self._lock):
            return super().accept(listener)

    def connect(self, conn, address):
        with QMutexLocker(self._lock):
            return super().connect(conn, address)


class _EventWorker(QThread):
    def __init__(
        self, proactor, event_poller: "_EventPoller", parent: "QObject" = None
    ):
        super().__init__(parent=parent)

        self.__proactor = proactor
        self.__sig_events = event_poller.sig_events
        self.__semaphore = QSemaphore()

    def start(self, **kwargs):
        super().start(**kwargs)
        self.__semaphore.acquire()

    def stop(self):
        self.requestInterruption()
        # Wait for thread to end
        self.wait()

    def run(self):
        log.debug("Thread started")
        self.__semaphore.release()

        while not self.isInterruptionRequested():
            events = self.__proactor.select(0.01)
            if events:
                log.debug("Got events from poll: %s", events)
                self.__sig_events.emit(events)

        log.debug("Exiting thread")


class _EventPoller:

    """Polling of events in separate thread."""

    def __init__(self, sig_events, rsc_parent: "QObject"):
        self.sig_events = sig_events
        self.__worker = None
        self._rsc_parent = rsc_parent

    def start(self, proactor):
        log.debug("Starting (proactor: %s)...", proactor)
        self.__worker = _EventWorker(proactor, self, parent=self._rsc_parent)
        self.__worker.start()

    def stop(self):
        log.debug("Stopping worker thread...")
        self.__worker.stop()



from typing import TYPE_CHECKING, Union, Optional, Callable

if TYPE_CHECKING:
    from qtasync._env import Signal, SignalInstance, QtCore  # noqa: F401

SIGNAL_TYPE = Union["Signal", "SignalInstance"]
MESSAGE_HANDLER_TYPE = Optional[
    Callable[["QtCore.QtMsgType", "QtCore.QMessageLogContext", str], None]
]



import sys
import logging
import functools
import asyncio

from qtasync._env import QObject, Signal, QCoreApplication, Slot, QTimerEvent
from qtasync.types.unbound import SIGNAL_TYPE
from qtasync.types.bound import PYTHON_TIME
from qtasync._util import qt_timeout

log = logging.getLogger(__name__)


def _make_signaller(*args):
    class Signaller(QObject):
        signal: SIGNAL_TYPE = Signal(*args)

    return Signaller()


class _SimpleTimer(QObject):
    def __init__(self, parent: "QObject"):
        super().__init__(parent=parent)
        self.__callbacks: dict[int, "asyncio.Handle"] = {}
        self._stopped = False
        self.__debug_enabled = False

    def add_callback(self, handle, delay: PYTHON_TIME = 0):
        timerid = self.startTimer(qt_timeout(delay))
        self.__log_debug("Registering timer id %s", timerid)
        assert timerid not in self.__callbacks
        self.__callbacks[timerid] = handle
        return handle

    def timerEvent(self, event: "QTimerEvent"):  # noqa: N802
        timerid = event.timerId()
        self.__log_debug("Timer event on id %s", timerid)
        if self._stopped:
            self.__log_debug("Timer stopped, killing %s", timerid)
            self.killTimer(timerid)
            del self.__callbacks[timerid]
        else:
            try:
                handle = self.__callbacks[timerid]
            except KeyError as e:
                self.__log_debug(e)
                pass
            else:
                if handle.cancelled():
                    self.__log_debug("Handle %s cancelled", handle)
                else:
                    self.__log_debug("Calling handle %s", handle)
                    handle._run()
            finally:
                del self.__callbacks[timerid]
                # handle = None
            self.killTimer(timerid)

    def stop(self):
        self.__log_debug("Stopping timers")
        self._stopped = True

    def set_debug(self, enabled):
        self.__debug_enabled = enabled

    def __log_debug(self, *args, **kwargs):
        if self.__debug_enabled:
            log.debug(*args, **kwargs)


def _fileno(fd):
    if isinstance(fd, int):
        return fd
    try:
        return int(fd.fileno())
    except (AttributeError, TypeError, ValueError):
        raise ValueError(f"Invalid file object: {fd!r}") from None


def asyncClose(fn):
    """Allow to run async code before application is closed."""

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        f = asyncio.ensure_future(fn(*args, **kwargs))
        while not f.done():
            QCoreApplication.instance().processEvents()

    return wrapper


def asyncSlot(*args):
    """Make a Qt async slot run on asyncio loop."""

    def _error_handler(task):
        try:
            task.result()
        except Exception:
            sys.excepthook(*sys.exc_info())

    def outer_decorator(fn):
        @Slot(*args)
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            task = asyncio.ensure_future(fn(*args, **kwargs))
            task.add_done_callback(_error_handler)
            return task

        return wrapper

    return outer_decorator



import logging
from typing import Union, Optional, Dict

from qtasync import get_timeout_compatibility_mode
from qtasync._env import QDeadlineTimer, QtCore

from .types.bound import QT_TIME, PYTHON_TIME
from .types.unbound import MESSAGE_HANDLER_TYPE


log = logging.getLogger(__name__)


def qt_timeout(time_secs: Union[float, PYTHON_TIME, None]) -> Optional[QT_TIME]:
    if time_secs is None:
        return None
    if get_timeout_compatibility_mode() and isinstance(time_secs, int):
        # If timeout compatibility mode is set, then integer timeouts are treated like Qt timeouts–durations measured
        # in milliseconds–and if it is a float then it is treated like a python time duration (seconds as a float)
        return QT_TIME(time_secs)
    else:
        return QT_TIME(time_secs * 1000)


def py_timeout(time_msecs: Union[int, QT_TIME, None]) -> Optional[PYTHON_TIME]:
    if time_msecs is None:
        return None
    return PYTHON_TIME(time_msecs / 1000.0)


def mk_q_deadline_timer(timeout: Optional[PYTHON_TIME]) -> "QDeadlineTimer":
    # TODO: Why doesn't ForeverConstant work?
    return (
        QDeadlineTimer(QDeadlineTimer.ForeverConstant.Forever)
        if timeout is None
        else QDeadlineTimer(qt_timeout(timeout))
    )


class QtLoggingMap:
    _log_level_map: Dict["QtCore.QtMsgType", int] = {
        QtCore.QtMsgType.QtDebugMsg: logging.DEBUG,
        QtCore.QtMsgType.QtInfoMsg: logging.INFO,
        QtCore.QtMsgType.QtWarningMsg: logging.WARNING,
        QtCore.QtMsgType.QtCriticalMsg: logging.ERROR,
        QtCore.QtMsgType.QtFatalMsg: logging.FATAL,
    }

    @classmethod
    def get_python_logging_level(cls, qt_log_level: "QtCore.QtMsgType") -> int:
        return cls._log_level_map[qt_log_level]


def qt_message_handler(
    msg_type: "QtCore.QtMsgType",
    context: "QtCore.QMessageLogContext",
    message: str,
    logger: "logging.Logger" = None,
):
    _log = logger or log
    py_log_level = QtLoggingMap.get_python_logging_level(msg_type)
    # pylint: disable=atakama-fstring-error
    _log.log(
        py_log_level,
        "#QT %s: %s (%s:%s, %s)",
        msg_type,
        message,
        context.file,
        context.line,
        context.file,
    )


def install_default_qt_message_handler():
    install_custom_qt_message_handler(qt_message_handler)


def install_custom_qt_message_handler(
    handler: MESSAGE_HANDLER_TYPE,
) -> Optional[Union[MESSAGE_HANDLER_TYPE, object]]:
    """
    :param handler: A new message handler function for all Qt log messages
    :return: The old message handling function if one was registered, else None. The return type is a union of object
        due to how the shiboken bindings were created, but the return value will not be an object and rather be None
        or a function.
    """
    return QtCore.qInstallMessageHandler(handler)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qtasync',
 'qtasync.qasyncio',
 'qtasync.qconcurrent',
 'qtasync.qthreading',
 'qtasync.types']

package_data = \
{'': ['*']}

extras_require = \
{'PyQt5': ['PyQt5>=5.15.6,<6.0.0', 'PyQt5-stubs>=5.15.2,<6.0.0'],
 'PyQt6': ['PyQt6>=6.2.2,<7.0.0'],
 'PySide2': ['PySide2==5.15.2'],
 'PySide6': ['PySide6>=6.2.2,<7.0.0']}

setup_kwargs = {
    'name': 'qtasync',
    'version': '0.3.1',
    'description': '',
    'long_description': None,
    'author': 'Brian Cefali',
    'author_email': 'brian@atakama.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import sys
import logging
import asyncio
import itertools
import time
from typing import Callable

from qtasync._env import QCoreApplication, QSocketNotifier, QObject
from qtasync.qconcurrent._futures import QtThreadPoolExecutor
from qtasync.types.unbound import SIGNAL_TYPE
from qtasync.types.bound import PYTHON_TIME
from qtasync.qasyncio._util import _SimpleTimer, _make_signaller, _fileno

log = logging.getLogger(__name__)


class _QEventLoop(asyncio.BaseEventLoop):
    def __init__(self, *args, **kwargs):
        self.__app = QCoreApplication.instance()
        assert self.__app is not None, "No QCoreApplication has been instantiated"
        self.__last_exit_code = None
        self.__is_running = False
        self.__debug_enabled = False
        self.__default_executor = None
        self.__exception_handler = None
        self._read_notifiers = {}
        self._write_notifiers = {}
        self._rsc_parent = QObject()
        self._timer = _SimpleTimer(parent=self._rsc_parent)

        self.__call_soon_signaller = signaller = _make_signaller(object, tuple)
        self.__call_soon_signal: SIGNAL_TYPE = signaller.signal
        signaller.signal.connect(
            lambda callback, _args: self.call_soon(callback, *_args)
        )
        super().__init__(*args, **kwargs)
        self.set_debug(True)

    def _before_run_forever(self):
        raise NotImplementedError

    def _after_run_forever(self):
        raise NotImplementedError

    def _check_closed(self):
        raise NotImplementedError

    def run_forever(self):
        """Run eventloop forever."""

        if self.__is_running:
            raise RuntimeError("Event loop already running")

        self.__is_running = True
        self._before_run_forever()

        try:
            self.__log_debug("Starting Qt event loop")
            asyncio.events._set_running_loop(self)  # noqa
            try:
                self.__last_exit_code = self.__app.exec_()
            except AttributeError:
                self.__last_exit_code = self.__app.exec()
            except:  # noqa: E722
                log.exception("Failed to run QCoreApplication event loop")
                self.__last_exit_code = -1
            self.__log_debug(
                "Qt event loop ended with result %s", self.__last_exit_code
            )
            return self.__last_exit_code
        finally:
            asyncio.events._set_running_loop(None)  # noqa
            self._after_run_forever()
            self.__is_running = False

    def run_until_complete(self, future):
        """Run until Future is complete."""

        if self.__is_running:
            raise RuntimeError("Event loop already running")

        self.__log_debug("Running %s until complete", future)
        future = asyncio.ensure_future(future, loop=self)

        def stop(*_args):
            self.stop()

        future.add_done_callback(stop)
        try:
            self.run_forever()
        finally:
            future.remove_done_callback(stop)
        self.__app.processEvents()  # run loop one last time to process all the events
        if not future.done():
            raise RuntimeError("Event loop stopped before Future completed.")

        self.__log_debug("Future %s finished running", future)
        return future.result()

    def stop(self):
        """Stop event loop."""
        if not self.__is_running:
            self.__log_debug("Already stopped")
            return

        self.__log_debug("Stopping event loop...")
        self.__is_running = False
        self.__app.exit()
        self.__log_debug("Stopped event loop")

    def is_running(self):
        """Return True if the event loop is running, False otherwise."""
        return self.__is_running

    def close(self):
        """
        Release all resources used by the event loop.

        The loop cannot be restarted after it has been closed.
        """
        if self.is_running():
            raise RuntimeError("Cannot close a running event loop")
        if self.is_closed():
            return

        self.__log_debug("Closing event loop...")
        if self.__default_executor is not None:
            self.__default_executor.shutdown()

        super().close()
        self._timer.stop()
        self._rsc_parent.deleteLater()
        self._timer = None
        self._rsc_parent = None

        for notifier in itertools.chain(
            self._read_notifiers.values(), self._write_notifiers.values()
        ):
            notifier.setEnabled(False)

        self._read_notifiers = None
        self._write_notifiers = None

    def call_later(self, delay: PYTHON_TIME, callback: Callable, *args, context=None):
        """Register callback to be invoked after a certain delay."""
        if asyncio.iscoroutinefunction(callback):
            raise TypeError("coroutines cannot be used with call_later")
        if not callable(callback):
            raise TypeError(
                "callback must be callable: {}".format(type(callback).__name__)
            )
        self._check_closed()

        self.__log_debug(
            "Registering callback %s to be invoked with arguments %s after %s second(s)",
            callback,
            args,
            delay,
        )

        return self._add_callback(
            asyncio.Handle(callback, args, self, context=context), delay
        )

    def _add_callback(self, handle: "asyncio.Handle", delay: PYTHON_TIME = 0):
        return self._timer.add_callback(handle, delay)

    def call_soon(self, callback: Callable, *args, context=None):
        """Register a callback to be run on the next iteration of the event loop."""
        return self.call_later(0, callback, *args, context=context)

    def call_at(self, when: PYTHON_TIME, callback: Callable, *args, context=None):
        """Register callback to be invoked at a certain time."""
        return self.call_later(when - self.time(), callback, *args, context=context)

    def time(self) -> PYTHON_TIME:
        """Get time according to event loop's clock."""
        return PYTHON_TIME(time.monotonic())

    def _add_reader(self, fd, callback, *args):
        """Register a callback for when a file descriptor is ready for reading."""
        self._check_closed()

        try:
            existing = self._read_notifiers[fd]
        except KeyError:
            pass
        else:
            # this is necessary to avoid race condition-like issues
            existing.setEnabled(False)
            existing.activated["int"].disconnect()
            # will get overwritten by the assignment below anyways

        notifier = QSocketNotifier(_fileno(fd), QSocketNotifier.Type.Read)
        notifier.setEnabled(True)
        self.__log_debug("Adding reader callback for file descriptor %s", fd)
        notifier.activated["int"].connect(
            lambda: self.__on_notifier_ready(
                self._read_notifiers, notifier, fd, callback, args
            )  # noqa: C812
        )
        self._read_notifiers[fd] = notifier

    def _remove_reader(self, fd):
        """Remove reader callback."""
        if self.is_closed():
            return

        self.__log_debug("Removing reader callback for file descriptor %s", fd)
        try:
            notifier = self._read_notifiers.pop(fd)
        except KeyError:
            return False
        else:
            notifier.setEnabled(False)
            return True

    def _add_writer(self, fd, callback, *args):
        """Register a callback for when a file descriptor is ready for writing."""
        self._check_closed()
        try:
            existing = self._write_notifiers[fd]
        except KeyError:
            pass
        else:
            # this is necessary to avoid race condition-like issues
            existing.setEnabled(False)
            existing.activated["int"].disconnect()
            # will get overwritten by the assignment below anyways

        notifier = QSocketNotifier(
            _fileno(fd),
            QSocketNotifier.Type.Write,
        )
        notifier.setEnabled(True)
        self.__log_debug("Adding writer callback for file descriptor %s", fd)
        notifier.activated["int"].connect(
            lambda: self.__on_notifier_ready(
                self._write_notifiers, notifier, fd, callback, args
            )  # noqa: C812
        )
        self._write_notifiers[fd] = notifier

    def _remove_writer(self, fd):
        """Remove writer callback."""
        if self.is_closed():
            return

        self.__log_debug("Removing writer callback for file descriptor %s", fd)
        try:
            notifier = self._write_notifiers.pop(fd)
        except KeyError:
            return False
        else:
            notifier.setEnabled(False)
            return True

    def __notifier_cb_wrapper(self, notifiers, notifier, fd, callback, args):
        # This wrapper gets called with a certain delay. We cannot know
        # for sure that the notifier is still the current notifier for
        # the fd.
        if notifiers.get(fd, None) is not notifier:
            return
        try:
            callback(*args)
        finally:
            # The notifier might have been overriden by the
            # callback. We must not re-enable it in that case.
            if notifiers.get(fd, None) is notifier:
                notifier.setEnabled(True)
            else:
                notifier.activated["int"].disconnect()

    def __on_notifier_ready(self, notifiers, notifier, fd, callback, args):
        if fd not in notifiers:
            log.warning(
                "Socket notifier for fd %s is ready, even though it should "
                "be disabled, not calling %s and disabling",
                fd,
                callback,
            )
            notifier.setEnabled(False)
            return

        # It can be necessary to disable QSocketNotifier when e.g. checking
        # ZeroMQ sockets for events
        assert notifier.isEnabled()
        self.__log_debug("Socket notifier for fd %s is ready", fd)
        notifier.setEnabled(False)
        self.call_soon(
            self.__notifier_cb_wrapper, notifiers, notifier, fd, callback, args
        )

    # Methods for interacting with threads.

    def call_soon_threadsafe(self, callback, *args, context=None):
        """Thread-safe version of call_soon."""
        self.__call_soon_signal.emit(callback, args)

    def run_in_executor(self, executor, callback, *args):
        """Run callback in executor.

        If no executor is provided, the default executor will be used, which defers execution to
        a background thread.
        """
        self.__log_debug("Running callback %s with args %s in executor", callback, args)
        if isinstance(callback, asyncio.Handle):
            assert not args
            assert not isinstance(callback, asyncio.TimerHandle)
            if callback.cancelled():
                f = asyncio.Future()
                f.set_result(None)
                return f
            callback, args = callback._callback, callback._args

        if executor is None:
            self.__log_debug("Using default executor")
            executor = self.__default_executor

        if executor is None:
            self.__log_debug("Creating default executor")
            executor = self.__default_executor = QtThreadPoolExecutor()

        return asyncio.wrap_future(executor.submit(callback, *args))

    def set_default_executor(self, executor):
        self.__default_executor = executor

    # Error handlers.

    def set_exception_handler(self, handler):
        log.info("Changing exception handler to %s", handler)
        self.__exception_handler = handler

    def default_exception_handler(self, context):
        """Handle exceptions.

        This is the default exception handler.

        This is called when an exception occurs and no exception
        handler is set, and can be called by a custom exception
        handler that wants to defer to the default behavior.

        context parameter has the same meaning as in
        `call_exception_handler()`.
        """
        self.__log_debug("Default exception handler executing")
        message = context.get("message")
        if not message:
            message = "Unhandled exception in event loop"

        try:
            exception = context["exception"]
        except KeyError:
            exc_info = False
        else:
            exc_info = (type(exception), exception, exception.__traceback__)

        log_lines = [message]
        for key in [k for k in sorted(context) if k not in {"message", "exception"}]:
            log_lines.append("{}: {!r}".format(key, context[key]))

        self.__log_error("\n".join(log_lines), exc_info=exc_info)

    def call_exception_handler(self, context):
        log.info("call exception handler, %s", context)
        if self.__exception_handler is None:
            try:
                self.default_exception_handler(context)
            except Exception:
                # Second protection layer for unexpected errors
                # in the default implementation, as well as for subclassed
                # event loops with overloaded "default_exception_handler".
                self.__log_error(
                    "Exception in default exception handler", exc_info=True
                )

            return

        try:
            self.__exception_handler(self, context)
        except Exception as exc:
            # Exception in the user set custom exception handler.
            try:
                # Let's try the default handler.
                self.default_exception_handler(
                    {
                        "message": "Unhandled error in custom exception handler",
                        "exception": exc,
                        "context": context,
                    }
                )
            except Exception:
                # Guard 'default_exception_handler' in case it's
                # overloaded.
                self.__log_error(
                    "Exception in default exception handler while handling an unexpected error "
                    "in custom exception handler",
                    exc_info=True,
                )

    # Debug flag management.

    def get_debug(self):
        return self.__debug_enabled

    def set_debug(self, enabled):
        super().set_debug(enabled)
        self.__debug_enabled = enabled
        self._timer.set_debug(enabled)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()
        self.close()

    def __log_debug(self, *args, **kwargs):
        if self.__debug_enabled:
            log.debug(*args, **kwargs)

    @classmethod
    def __log_error(cls, *args, **kwds):
        # In some cases, the error method itself fails, don't have a lot of options in that case
        try:
            log.error(*args, **kwds)
        except:  # noqa E722
            sys.stderr.write("{!r}, {!r}\n".format(args, kwds))



_timeout_compatibility_mode = False


def set_timeout_compatibility_mode(compat_mode: bool):
    global _timeout_compatibility_mode
    _timeout_compatibility_mode = compat_mode


def get_timeout_compatibility_mode() -> bool:
    global _timeout_compatibility_mode
    return _timeout_compatibility_mode



from ._timer import QtTimer
from ._locks import (
    QtLock,
    QtRLock,
    QtCondition,
    QtEvent,
    QtSemaphore,
)
from ._thread import QtThread


__all__ = [
    "QtTimer",
    "QtLock",
    "QtRLock",
    "QtCondition",
    "QtEvent",
    "QtThread",
    "QtSemaphore",
]
